
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "FFT.h"

extern void	Error();

static void	Abort(path,msg)
char		*path,*msg;
{
	(void)fprintf(stderr,"%s : ",path); Error(msg);
}

#define BYTES(type,size)	(sizeof(type)*(size_t)(size))
#define ALLOC(type,size)	(type *)malloc(BYTES(type,size))

static int	Scan(path,data)
char		*path;
double		**data;
{
	FILE	*file;
	int	N,x;

	if (*path=='-')
	    file=stdin;
	else
	    if ((file=fopen(path,"r"))==NULL) Abort(path,"file not found.");

	if (fscanf(file,"%d",&N)!=1) Abort(path,"bad number of data.");

	if ((*data=ALLOC(double,N))==NULL)
	    Abort(path,"data memory allocation error.");

	for (x=0; x<N; x++)
	    if (fscanf(file,"%lf",*data+x)!=1) Abort(path,"data read error.");

	if (file!=stdin) (void)fclose(file);

	return N;
}

static Complex	Product(r1,i1,r2,i2)
double		r1,i1,r2,i2;
{
	Complex	c;

	c.r=r1*r2-i1*i2;
	c.i=r1*i2+i1*r2; return c;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Ng,Nh,D1,D2,ng,nh,xg,xh,D0,L,L2,x,Ng1,Nh1,L1,g,h,Ngh;
	double	*G,*H,f,t,r,i,w,msd;
	Complex	*E,*e,*P,*p,*Q,*q,*a,*b,g2,h2,gh;
	char	*mae="memory allocation error.";

	if (argc!=3 && argc!=5) Error("usage : rmsd_1d data1 data2 {D1 D2}");

	Ng=Scan(argv[1],&G);
	Nh=Scan(argv[2],&H);

	if (argc==3) {
	    D1=1-Nh; D2=Ng-1;
	}
	else
	    if ((D1=atoi(argv[3]))>
		(D2=atoi(argv[4])) || D1<=(-Nh) || D2>=Ng)
		Error("bad range of distance.");

	if ((ng=Nh+D2)>Ng) ng=Ng;
	if ((nh=Ng-D1)>Nh) nh=Nh;

	if (D2<0) {
	    xg=0; xh=(-(D0=D2));
	}
	else if (D1>0) {
	    xg=D0=D1; xh=0;
	}
	else
	    xg=xh=D0=0;

	Ng=ng-xg;
	Nh=nh-xh; D1-=D0; D2-=D0;

	if ((L=SetUpFFT(((Ng>Nh)?Ng:Nh)+
			((D2<0)?1-D1:(D1>0)?D2+1:(-D1>D2)?1-D1:D2+1),&E))==0 ||
	    (E=e=(Complex *)realloc(E,BYTES(Complex,L2=L*2)))==NULL ||
	    (P=p=ALLOC(Complex,L))==NULL ||
	    (Q=q=ALLOC(Complex,L))==NULL) Error(mae);

	for (x=0; x<L; x++, e++, p++, q++) {
	    e[L].r=(-e->r);
	    e[L].i=(-e->i);

	    if (x<Ng) {
		f=G[xg+x]; p->r=f*f; q->r=f;
	    }
	    else
		p->r=q->r=0.0;

	    if (x<Nh) {
		f=H[xh+x]; p->i=f*f; q->i=f;
	    }
	    else
		p->i=q->i=0.0;
	}
	FFT(L,E,P,-1);
	FFT(L,E,Q,-1);

	Ng1=Ng-1;
	Nh1=Nh-1; L1=L-1;

	P->r=P->r*(double)Nh+(double)Ng*P->i-2.0*Q->r*Q->i;
	P->i=0.0;
	for (e=E+1, g=Ng1, h=Nh1, p=P+1, a=P+L1, q=Q+1, b=Q+L1;
	     p<a;
	     e++, g=(g+Ng1)%L2, h=(h+Nh1)%L2, p++, a--, q++, b--) {
	    t=e->r/e->i;
	    r=E[h].r;
	    i=E[h].i; w=r+i*t; g2=Product(p->r+a->r,p->i-a->i,w*r, w*i);
	    r=E[g].r;
	    i=E[g].i; w=r+i*t; h2=Product(w*r,-w*i,p->i+a->i,p->r-a->r);

	    gh=Product(q->r+b->r,q->i-b->i,q->i+b->i,q->r-b->r);

	    p->r=
	    a->r=(g2.r+h2.r-gh.r)/2.0;
	    p->i=(g2.i+h2.i-gh.i)/2.0;
	    a->i=(-p->i);
	}
	p->r=p->r*(double)(Nh%2)+(double)(Ng%2)*p->i-2.0*q->r*q->i;
	p->i=0.0;

	FFT(L,E,P, 1);

	Ngh=Ng-Nh;

	for (x=D1; x<=D2; x++) {
	    if (x<0)
		msd=P[L+x].r/(double)((x>=Ngh)?Ng:Nh+x);
	    else
		msd=P[x].r/(double)((x>=Ngh)?Ng-x:Nh);

	    (void)printf("%d\t%lf\n",x+D0,sqrt(((msd<0.0)?0.0:msd)/(double)L));
	}
	return 0;
}
