
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "FFT.h"

#ifndef	CSI_BS
#define CSI_BS
#endif

#include "csi.h"

extern void	Error();

static void	Trim(Ng,Nh,D1,D2,Og,Oh,D0)
int		*Ng,*Nh,*D1,*D2,*Og,*Oh,*D0;
{
	int	ng,nh;

	if ((ng=(*Nh)+(*D2))>(*Ng)) ng=(*Ng);
	if ((nh=(*Ng)-(*D1))>(*Nh)) nh=(*Nh);

	if (*D2<0) {
	    *Og=0; *Oh=(-((*D0)=(*D2)));
	}
	else if (*D1>0) {
	    *Og=(*D0)=(*D1); *Oh=0;
	}
	else
	    *Og=(*Oh)=(*D0)=0;

	*Ng=ng-(*Og);
	*Nh=nh-(*Oh); *D1-=(*D0); *D2-=(*D0);
}

#define BYTES(type,size)	(sizeof(type)*(size_t)(size))

static Complex	*Expand(L,E)
int		L;
Complex		*E;
{
	Complex	*e;
	int	l;

	if ((E=(Complex *)realloc(E,BYTES(Complex,L*2)))!=NULL)
	    for (e=E, l=0; l<L; l++, e++) {
		e[L].r=(-e->r);
		e[L].i=(-e->i);
	    }

	return E;
}

static void	Read(pbs,Ox,Oy,Nx,Ny,data)
PBS		*pbs;
int		Ox,Oy,Nx,Ny;
double		**data;
{
	int	Mx,My,y,x;
	Cell	**cell;

	ReadImageFile(pbs->path,&Mx,&My,NULL,&cell,NULL);

	Nx+=Ox;
	Ny+=Oy;

	for (y=My-1; y>=Ny; y--) free(cell[y]);

	for (y=Ny-1; y>=My; y--)
	    for (x=Nx-1; x>=Ox; x--) data[y-Oy][x-Ox]=pbs->base;

	for (; y>=Oy; y--) {
	    for (x=Nx-1; x>=Mx; x--) data[y-Oy][x-Ox]=pbs->base;

	    for (; x>=Ox; x--)
		data[y-Oy][x-Ox]=pbs->base+pbs->step*(double)cell[y][x];

	    free(cell[y]);
	}
	for (; y>=0; y--) free(cell[y]);

	free(cell);
}

#define PATH_LEN	1023

static void	Dump(tmp,Lx,Ly,z,data)
char		*tmp;
int		Lx,Ly,z;
Complex		**data;
{
	char	path[PATH_LEN];
        FILE	*file;
        int	y;

	(void)sprintf(path,tmp,z);
	if ((file=fopen(path,"wb"))==NULL)
	    Error("can not create temporary file.");

	for (y=0; y<Ly; y++)
	    if (fwrite(*(data++),sizeof(Complex),(size_t)Lx,file)!=Lx)
		Error("can not write temporary data.");

	(void)fclose(file);
}

static char	*cnotf="can not open temporary file.",
		*cnstd="can not scan temporary data.";

static void	Collect1(tmp,Lx,y,Lz,data)
char		*tmp;
int		Lx,y,Lz;
Complex		**data;
{
        char	path[PATH_LEN];
        FILE	*file;
        int	z;
	long	offset=sizeof(Complex)*(long)Lx*(long)y;

	for (z=0; z<Lz; z++) {
	    (void)sprintf(path,tmp,z);
	    if ((file=fopen(path,"rb"))==NULL) Error(cnotf);

	    if (fseek(file,offset,SEEK_SET) ||
		fread(*(data++),sizeof(Complex),(size_t)Lx,file)!=Lx)
		Error(cnstd);

	    (void)fclose(file);
	}
}

static void	Collect2(tmp,Lx,Y1,Y2,Lz,data1,data2)
char		*tmp;
int		Lx,Y1,Y2,Lz;
Complex		**data1,**data2;
{
        char	path[PATH_LEN];
        FILE	*file;
        int	z;
	long	size=sizeof(Complex)*(long)Lx,offset1=size*(long)Y1,
					      offset2=size*(long)Y2;

	for (z=0; z<Lz; z++) {
	    (void)sprintf(path,tmp,z);
	    if ((file=fopen(path,"rb"))==NULL) Error(cnotf);

	    if (fseek(file,offset1,SEEK_SET) ||
		fread(*(data1++),sizeof(Complex),(size_t)Lx,file)!=Lx ||
		fseek(file,offset2,SEEK_SET) ||
		fread(*(data2++),sizeof(Complex),(size_t)Lx,file)!=Lx)
		Error(cnstd);

	    (void)fclose(file);
	}
}

static char	*cnrtf="can not reopen temporary file.",
		*cnrtd="can not replace temporary data.";

static void	Replace1(tmp,Lx,y,Lz,data)
char		*tmp;
int		Lx,y,Lz;
Complex		**data;
{
	char	path[PATH_LEN];
	FILE	*file;
	int	z;
	long	offset=sizeof(Complex)*(long)Lx*(long)y;

	for (z=0; z<Lz; z++) {
	    (void)sprintf(path,tmp,z);
	    if ((file=fopen(path,"rb+"))==NULL) Error(cnrtf);

	    if (fseek(file,offset,SEEK_SET) ||
		fwrite(*(data++),sizeof(Complex),(size_t)Lx,file)!=Lx)
		Error(cnrtd);

	    (void)fclose(file);
	}
}

static void	Replace2(tmp,Lx,Y1,Y2,Lz,data1,data2)
char		*tmp;
int		Lx,Y1,Y2,Lz;
Complex		**data1,**data2;
{
	char	path[PATH_LEN];
	FILE	*file;
	int	z;
	long	size=sizeof(Complex)*(long)Lx, offset1=size*(long)Y1,
					       offset2=size*(long)Y2;

	for (z=0; z<Lz; z++) {
	    (void)sprintf(path,tmp,z);
	    if ((file=fopen(path,"rb+"))==NULL) Error(cnrtf);

	    if (fseek(file,offset1,SEEK_SET) ||
		fwrite(*(data1++),sizeof(Complex),(size_t)Lx,file)!=Lx ||
		fseek(file,offset2,SEEK_SET) ||
		fwrite(*(data2++),sizeof(Complex),(size_t)Lx,file)!=Lx)
		Error(cnrtd);

	    (void)fclose(file);
	}
}

static void	Restore(tmp,Lx,Ly,z,data)
char		*tmp;
int		Lx,Ly,z;
Complex		**data;
{
	char	path[PATH_LEN];
	FILE	*file;
	int	y;

	(void)sprintf(path,tmp,z);
	if ((file=fopen(path,"rb"))==NULL) Error(cnotf);

	for (y=0; y<Ly; y++)
	    if (fread(*(data++),sizeof(Complex),(size_t)Lx,file)!=Lx)
		Error(cnstd);

	(void)fclose(file);
}

static Complex	Product(r1,i1,r2,i2)
double		r1,i1,r2,i2;
{
	Complex	c;

	c.r=r1*r2-i1*i2;
	c.i=r1*i2+i1*r2; return c;
}

#define ALLOC(type,size)	(type *)malloc(BYTES(type,size))

int	main(argc,argv)
int	argc;
char	**argv;
{
	PBS	*PBSg,*PBSh;
	int	Ngx,Ngy,Ngz,Nhx,Nhy,Nhz,Dx1,Dy1,Dz1,Dx2,Dy2,Dz2,
		xg,yg,zg,xh,yh,zh,Dx0,Dy0,Dz0,z,x,y,Lx,Ly,Lz,len,
		Lx1,Ly1,Lz1,Lx2,Ly2,Lz2,Ngx1,Ngy1,Ngz1,Nhx1,Nhy1,Nhz1,
		gz,hz,gx,hx,gy,hy,Lxx,Lyy,Nghx,Nghy,Nghz;
	double	**F,**Fy,*Fx,f,Ngx2,Ngy2,Ngz2,Nhx2,Nhy2,Nhz2,
		Ngxy,Nhxy,Ngx2y,Nhx2y,Ngxz,Nhxz,Ngxy2,Nhxy2,Ngx2y2,Nhx2y2,
		t,r,i,w,rg,ig,rh,ih,L,LNz,LNyz,msd;
	Complex	*Ex,*Ey,*Ez,*U,*V,*A,*B,**P,**Pyz,**Q,**Qyz,**R,**Rz,**S,**Sz,
		*Px,*Qx,*u,*v,*a,*ez,*b,g2,h2,gh,*ex,c,*ey,cg,ch;
	char	tmp1[PATH_LEN],tmp2[PATH_LEN],
		*mae="memory allocation error.";

	if (argc!=6 && argc!=7 && argc!=12 && argc!=13)
	    Error(
"usage : rmsd_3d.VM dir1 NBS1 dir2 NBS2 {Dx1 Dy1 Dz1 Dx2 Dy2 Dz2} tmp1 {tmp2}"
	    );

	PBSg=CheckSliceImages(argv[1],argv[2],&Ngx,&Ngy,&Ngz,NULL);
	PBSh=CheckSliceImages(argv[3],argv[4],&Nhx,&Nhy,&Nhz,NULL);

	if (argc==6 || argc==7) {
	    Dx1=1-Nhx; Dx2=Ngx-1;
	    Dy1=1-Nhy; Dy2=Ngy-1;
	    Dz1=1-Nhz; Dz2=Ngz-1;
	}
	else
	    if ((Dx1=atoi(argv[ 5]))>
		(Dx2=atoi(argv[ 8])) || Dx1<=(-Nhx) || Dx2>=Ngx ||
		(Dy1=atoi(argv[ 6]))>
		(Dy2=atoi(argv[ 9])) || Dy1<=(-Nhy) || Dy2>=Ngy ||
		(Dz1=atoi(argv[ 7]))>
		(Dz2=atoi(argv[10])) || Dz1<=(-Nhz) || Dz2>=Ngz)
		Error("bad range of distance.");

	Trim(&Ngx,&Nhx,&Dx1,&Dx2,&xg,&xh,&Dx0);
	Trim(&Ngy,&Nhy,&Dy1,&Dy2,&yg,&yh,&Dy0);
	Trim(&Ngz,&Nhz,&Dz1,&Dz2,&zg,&zh,&Dz0);

	z=(Ngy>Nhy)?Ngy:Nhy;

	if ((F=Fy=ALLOC(double *,z))==NULL) Error(mae);

	x=(Ngx>Nhx)?Ngx:Nhx;

	for (y=0; y<z; y++) if ((*(Fy++)=ALLOC(double,x))==NULL) Error(mae);

#define SETUPFFT(Ng,Nh,D1,D2,E)	\
	SetUpFFT(((Ng>Nh)?Ng:Nh)+ \
		 ((D2<0)?1-D1:(D1>0)?D2+1:(-D1>D2)?1-D1:D2+1),&E)

	if ((Lx=SETUPFFT(Ngx,Nhx,Dx1,Dx2,Ex))==0 ||
	    (Ex=Expand(Lx,Ex))==NULL ||
	    (Ly=SETUPFFT(Ngy,Nhy,Dy1,Dy2,Ey))==0 ||
	    (Ey=Expand(Ly,Ey))==NULL ||
	    (Lz=SETUPFFT(Ngz,Nhz,Dz1,Dz2,Ez))==0 ||
	    (Ez=Expand(Lz,Ez))==NULL) Error(mae);

	z=(Ly>Lz)?Ly:Lz;

	if ((U=ALLOC(Complex,z))==NULL ||
	    (V=ALLOC(Complex,z))==NULL ||
	    (A=ALLOC(Complex,Lz))==NULL ||
	    (B=ALLOC(Complex,Lz))==NULL ||
	    (P=Pyz=ALLOC(Complex *,z))==NULL ||
	    (Q=Qyz=ALLOC(Complex *,z))==NULL ||
	    (R=Rz=ALLOC(Complex *,Lz))==NULL ||
	    (S=Sz=ALLOC(Complex *,Lz))==NULL) Error(mae);

	for (y=0; y<z; y++)
	    if ((*(Pyz++)=ALLOC(Complex,Lx))==NULL ||
		(*(Qyz++)=ALLOC(Complex,Lx))==NULL) Error(mae);

	for (z=0; z<Lz; z++)
	    if ((*(Rz++)=ALLOC(Complex,Lx))==NULL ||
		(*(Sz++)=ALLOC(Complex,Lx))==NULL) Error(mae);

	len=1; for (z=Lz-1; z>=10; z/=10) ++len;
	(void)sprintf(tmp1,"%s/1%%0%dd",argv[argc-((argc==6 || argc==12)?1:2)],
					len);
	(void)sprintf(tmp2,"%s/2%%0%dd",argv[argc-1],
					len);

	for (z=0; z<Lz; z++) {
	    if (z<Ngz) {
		Read(PBSg+z+zg,xg,yg,Ngx,Ngy,F);

		for (Fy=F, Pyz=P, Qyz=Q, y=0; y<Ly; y++, Pyz++, Qyz++)
		    if (y<Ngy)
			for (Fx=(*(Fy++)),
			     Px=(*Pyz), Qx=(*Qyz), x=0; x<Lx; x++, Px++, Qx++)
			    if (x<Ngx) {
				f=(*(Fx++)); Px->r=f*f; Qx->r=f;
			    }
			    else
				Px->r=Qx->r=0.0;
		    else
			for (Px=(*Pyz), Qx=(*Qyz), x=0; x<Lx; x++, Px++, Qx++)
			    Px->r=Qx->r=0.0;
	    }
	    else
		for (Pyz=P, Qyz=Q, y=0; y<Ly; y++, Pyz++, Qyz++)
		    for (Px=(*Pyz), Qx=(*Qyz), x=0; x<Lx; x++, Px++, Qx++)
			Px->r=Qx->r=0.0;

	    if (z<Nhz) {
		Read(PBSh+z+zh,xh,yh,Nhx,Nhy,F);

		for (Fy=F, Pyz=P, Qyz=Q, y=0; y<Ly; y++, Pyz++, Qyz++)
		    if (y<Nhy)
			for (Fx=(*(Fy++)),
			     Px=(*Pyz), Qx=(*Qyz), x=0; x<Lx; x++, Px++, Qx++)
			    if (x<Nhx) {
				f=(*(Fx++)); Px->i=f*f; Qx->i=f;
			    }
			    else
				Px->i=Qx->i=0.0;
		    else
			for (Px=(*Pyz), Qx=(*Qyz), x=0; x<Lx; x++, Px++, Qx++)
			    Px->i=Qx->i=0.0;
	    }
	    else
		for (Pyz=P, Qyz=Q, y=0; y<Ly; y++, Pyz++, Qyz++)
		    for (Px=(*Pyz), Qx=(*Qyz), x=0; x<Lx; x++, Px++, Qx++)
			Px->i=Qx->i=0.0;

	    if (z<Ngz || z<Nhz) {
		for (Pyz=P, Qyz=Q, y=0; y<Ngy || y<Nhy; y++) {
		    FFT(Lx,Ex,*(Pyz++),-1);
		    FFT(Lx,Ex,*(Qyz++),-1);
		}
		for (x=0; x<Lx; x++) {
		    for (u=U, v=V, y=0; y<Ly; y++) {
			*(u++)=P[y][x];
			*(v++)=Q[y][x];
		    }
		    FFT(Ly,Ey,U,-1);
		    FFT(Ly,Ey,V,-1);

		    for (u=U, v=V, y=0; y<Ly; y++) {
			P[y][x]=(*(u++));
			Q[y][x]=(*(v++));
		    }
		}
	    }
	    Dump(tmp1,Lx,Ly,z,P);
	    Dump(tmp2,Lx,Ly,z,Q);
	}
	Lx1=Lx-1; Ly1=Ly-1; Lz1=Lz-1;

	Lx2=Lx*2; Ly2=Ly*2; Lz2=Lz*2;
	Ngx1=Ngx-1; Ngy1=Ngy-1; Ngz1=Ngz-1;
	Nhx1=Nhx-1; Nhy1=Nhy-1; Nhz1=Nhz-1;
	Ngx2=(double)(Ngx%2); Ngy2=(double)(Ngy%2); Ngz2=(double)(Ngz%2);
	Nhx2=(double)(Nhx%2); Nhy2=(double)(Nhy%2); Nhz2=(double)(Nhz%2);

	Ngxy=(double)Ngx*(double)Ngy; Nhxy=(double)Nhx*(double)Nhy;
	Ngx2y=Ngx2*(double)Ngy; Nhx2y=Nhx2*(double)Nhy;
	Ngxz=(double)Ngx*(double)Ngz; Nhxz=(double)Nhx*(double)Nhz;
	Ngxy2=(double)Ngx*Ngy2; Nhxy2=(double)Nhx*Nhy2;
	Ngx2y2=Ngx2*Ngy2; Nhx2y2=Nhx2*Nhy2;
/*
    y=0 & x=0
*/
	Collect1(tmp1,Lx,0,Lz,P);
	Collect1(tmp2,Lx,0,Lz,R);

	for (u=U, a=A, z=0; z<Lz; z++) {
	    *(u++)=P[z][0]; *(a++)=R[z][0];
	}
	FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);

	U->r=U->r*Nhxy*(double)Nhz+Ngxy*(double)Ngz*U->i-2.0*A->r*A->i;
	U->i=0.0;
	for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=U+Lz1,
					a=A+1, b=A+Lz1;
	     u<v;
	     ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
						       a++, b--) {
	    t=ez->r/ez->i;
	    r=Ez[hz].r;
	    i=Ez[hz].i; w=Nhxy*(r+i*t);
			g2=Product(u->r+v->r,u->i-v->i,w*r, w*i);
	    r=Ez[gz].r;
	    i=Ez[gz].i; w=Ngxy*(r+i*t);
			h2=Product(w*r,-w*i,u->i+v->i,u->r-v->r);

	    gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

	    v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		 u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*Nhxy*Nhz2+Ngxy*Ngz2*u->i-2.0*a->r*a->i;
	u->i=0.0;

	FFT(Lz,Ez,U, 1);

	for (u=U, z=0; z<Lz; z++) P[z][0]=(*(u++));
/*
    y=0 & x=1~Lx/2-1
*/
	for (ex=Ex+1, gx=Ngx1, hx=Nhx1, x=1, Lxx=Lx1;
	     x<Lxx;
	     ex++, gx=(gx+Ngx1)%Lx2, hx=(hx+Nhx1)%Lx2, x++, Lxx--) {
	    t=ex->r/ex->i;
	    r=Ex[hx].r;
	    i=Ex[hx].i; w=(double)Nhy*(r+i*t); rh=w*r; ih=w*i;
	    r=Ex[gx].r;
	    i=Ex[gx].i; w=(double)Ngy*(r+i*t); rg=w*r; ig=w*i;

	    for (u=U, v=V, a=A, b=B, z=0; z<Lz; z++) {
		*(u++)=P[z][  x]; *(a++)=R[z][  x];
		*(v++)=P[z][Lxx]; *(b++)=R[z][Lxx];
	    }
	    FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);
	    FFT(Lz,Ez,V,-1); FFT(Lz,Ez,B,-1);

	    g2=Product(U->r+V->r,U->i-V->i,rh*(double)Nhz, ih*(double)Nhz);
	    h2=Product(rg*(double)Ngz,-ig*(double)Ngz,U->i+V->i,U->r-V->r);
	    gh=Product(A->r+B->r,A->i-B->i,A->i+B->i,A->r-B->r);

	    V->r=U->r=(g2.r+h2.r-gh.r)/2.0;
		 U->i=(g2.i+h2.i-gh.i)/2.0; V->i=(-U->i);
	    for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=V+Lz1,
					    a=A+1, b=B+Lz1, z=1;
		 z<Lz;
		 ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
							   a++, b--, z++) {
		t=ez->r/ez->i;
		r=Ez[hz].r;
		i=Ez[hz].i; w=r+i*t; c=Product(rh,ih,w*r,w*i);
				     g2=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		r=Ez[gz].r;
		i=Ez[gz].i; w=r+i*t; c=Product(rg,ig,w*r,w*i);
				     h2=Product(c.r,-c.i,u->i+v->i,u->r-v->r);

		gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

		v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		     u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	    }
	    FFT(Lz,Ez,U, 1);
	    FFT(Lz,Ez,V, 1);

	    for (u=U, v=V, z=0; z<Lz; z++) {
		P[z][  x]=(*(u++));
		P[z][Lxx]=(*(v++));
	    }
	}
/*
    y=0 & x=Lx/2
*/
	for (u=U, a=A, z=0; z<Lz; z++) {
	    *(u++)=P[z][x]; *(a++)=R[z][x];
	}
	FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);

	U->r=U->r*Nhx2y*(double)Nhz+Ngx2y*(double)Ngz*U->i-2.0*A->r*A->i;
	U->i=0.0;
	for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=U+Lz1,
					a=A+1, b=A+Lz1;
	     u<v;
	     ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
						       a++, b--) {
	    t=ez->r/ez->i;
	    r=Ez[hz].r;
	    i=Ez[hz].i; w=Nhx2y*(r+i*t);
			g2=Product(u->r+v->r,u->i-v->i,w*r, w*i);
	    r=Ez[gz].r;
	    i=Ez[gz].i; w=Ngx2y*(r+i*t);
			h2=Product(w*r,-w*i,u->i+v->i,u->r-v->r);

	    gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

	    v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		 u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*Nhx2y*Nhz2+Ngx2y*Ngz2*u->i-2.0*a->r*a->i;
	u->i=0.0;

	FFT(Lz,Ez,U, 1);

	for (u=U, z=0; z<Lz; z++) P[z][x]=(*(u++));

	Replace1(tmp1,Lx,0,Lz,P);
/*
    y=1~Ly/2-1 & x=0
*/
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, y=1, Lyy=Ly1;
	     y<Lyy;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, y++, Lyy--) {
	    t=ey->r/ey->i;
	    r=Ey[hy].r;
	    i=Ey[hy].i; w=r+i*t; rh=w*r; ih=w*i;
	    r=Ey[gy].r;
	    i=Ey[gy].i; w=r+i*t; rg=w*r; ig=w*i;

	    Collect2(tmp1,Lx,y,Lyy,Lz,P,Q);
	    Collect2(tmp2,Lx,y,Lyy,Lz,R,S);

	    for (u=U, v=V, a=A, b=B, z=0; z<Lz; z++) {
		*(u++)=P[z][0]; *(a++)=R[z][0];
		*(v++)=Q[z][0]; *(b++)=S[z][0];
	    }
	    FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);
	    FFT(Lz,Ez,V,-1); FFT(Lz,Ez,B,-1);

	    g2=Product(U->r+V->r,U->i-V->i,rh*Nhxz, ih*Nhxz);
	    h2=Product(rg*Ngxz,-ig*Ngxz,U->i+V->i,U->r-V->r);
	    gh=Product(A->r+B->r,A->i-B->i,A->i+B->i,A->r-B->r);

	    V->r=U->r=(g2.r+h2.r-gh.r)/2.0;
		 U->i=(g2.i+h2.i-gh.i)/2.0; V->i=(-U->i);
	    for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=V+Lz1,
					    a=A+1, b=B+Lz1, z=1;
		 z<Lz;
		 ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
							   a++, b--, z++) {
		t=ez->r/ez->i;
		r=Ez[hz].r;
		i=Ez[hz].i; w=(double)Nhx*(r+i*t);
			    c=Product(rh,ih,w*r,w*i);
			    g2=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		r=Ez[gz].r;
		i=Ez[gz].i; w=(double)Ngx*(r+i*t);
			    c=Product(rg,ig,w*r,w*i);
			    h2=Product(c.r,-c.i,u->i+v->i,u->r-v->r);

		gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

		v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		     u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	    }
	    FFT(Lz,Ez,U, 1);
	    FFT(Lz,Ez,V, 1);

	    for (u=U, v=V, z=0; z<Lz; z++) {
		P[z][0]=(*(u++));
		Q[z][0]=(*(v++));
	    }
/*
    y=1~Ly/2-1 & x=1~Lx-1
*/
	    for (ex=Ex+1, gx=Ngx1, hx=Nhx1, x=1, Lxx=Lx1;
		 x<Lx;
		 ex++, gx=(gx+Ngx1)%Lx2, hx=(hx+Nhx1)%Lx2, x++, Lxx--) {
		t=ex->r/ex->i;
		r=Ex[hx].r;
		i=Ex[hx].i; w=r+i*t; ch=Product(rh,ih,w*r,w*i);
		r=Ex[gx].r;
		i=Ex[gx].i; w=r+i*t; cg=Product(rg,ig,w*r,w*i);

		for (u=U, v=V, a=A, b=B, z=0; z<Lz; z++) {
		    *(u++)=P[z][  x]; *(a++)=R[z][  x];
		    *(v++)=Q[z][Lxx]; *(b++)=S[z][Lxx];
		}
		FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);
		FFT(Lz,Ez,V,-1); FFT(Lz,Ez,B,-1);

		g2=Product(U->r+V->r,U->i-V->i, ch.r*(double)Nhz,
						ch.i*(double)Nhz);
		h2=Product( cg.r*(double)Ngz,
			   -cg.i*(double)Ngz,U->i+V->i,U->r-V->r);
		gh=Product(A->r+B->r,A->i-B->i,A->i+B->i,A->r-B->r);

		V->r=U->r=(g2.r+h2.r-gh.r)/2.0;
		     U->i=(g2.i+h2.i-gh.i)/2.0; V->i=(-U->i);
		for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=V+Lz1,
						a=A+1, b=B+Lz1, z=1;
		     z<Lz;
		     ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
							       a++, b--, z++) {
		    t=ez->r/ez->i;
		    r=Ez[hz].r;
		    i=Ez[hz].i; w=r+i*t;
				c=Product(ch.r,ch.i,w*r,w*i);
				g2=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		    r=Ez[gz].r;
		    i=Ez[gz].i; w=r+i*t;
				c=Product(cg.r,cg.i,w*r,w*i);
				h2=Product(c.r,-c.i,u->i+v->i,u->r-v->r);

		    gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

		    v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
			 u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
		}
		FFT(Lz,Ez,U, 1);
		FFT(Lz,Ez,V, 1);

		for (u=U, v=V, z=0; z<Lz; z++) {
		    P[z][  x]=(*(u++));
		    Q[z][Lxx]=(*(v++));
		}
	    }
	    Replace2(tmp1,Lx,y,Lyy,Lz,P,Q);
	}
/*
    y=Ly/2 & x=0
*/
	Collect1(tmp1,Lx,y,Lz,P);
	Collect1(tmp2,Lx,y,Lz,R);

	for (u=U, a=A, z=0; z<Lz; z++) {
	    *(u++)=P[z][0]; *(a++)=R[z][0];
	}
	FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);

	U->r=U->r*Nhxy2*(double)Nhz+Ngxy2*(double)Ngz*U->i-2.0*A->r*A->i;
	U->i=0.0;
	for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=U+Lz1,
					a=A+1, b=A+Lz1;
	     u<v;
	     ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
						       a++, b--) {
	    t=ez->r/ez->i;
	    r=Ez[hz].r;
	    i=Ez[hz].i; w=Nhxy2*(r+i*t);
			g2=Product(u->r+v->r,u->i-v->i,w*r, w*i);
	    r=Ez[gz].r;
	    i=Ez[gz].i; w=Ngxy2*(r+i*t);
			h2=Product(w*r,-w*i,u->i+v->i,u->r-v->r);

	    gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

	    v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		 u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*Nhxy2*Nhz2+Ngxy2*Ngz2*u->i-2.0*a->r*a->i;
	u->i=0.0;

	FFT(Lz,Ez,U, 1);

	for (u=U, z=0; z<Lz; z++) P[z][0]=(*(u++));
/*
    y=Ly/2 & x=1~Lx/2-1
*/
	for (ex=Ex+1, gx=Ngx1, hx=Nhx1, x=1, Lxx=Lx1;
	     x<Lxx;
	     ex++, gx=(gx+Ngx1)%Lx2, hx=(hx+Nhx1)%Lx2, x++, Lxx--) {
	    t=ex->r/ex->i;
	    r=Ex[hx].r;
	    i=Ex[hx].i; w=Nhy2*(r+i*t); rh=w*r; ih=w*i;
	    r=Ex[gx].r;
	    i=Ex[gx].i; w=Ngy2*(r+i*t); rg=w*r; ig=w*i;

	    for (u=U, v=V, a=A, b=B, z=0; z<Lz; z++) {
		*(u++)=P[z][  x]; *(a++)=R[z][  x];
		*(v++)=P[z][Lxx]; *(b++)=R[z][Lxx];
	    }
	    FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);
	    FFT(Lz,Ez,V,-1); FFT(Lz,Ez,B,-1);

	    g2=Product(U->r+V->r,U->i-V->i,rh*(double)Nhz, ih*(double)Nhz);
	    h2=Product(rg*(double)Ngz,-ig*(double)Ngz,U->i+V->i,U->r-V->r);
	    gh=Product(A->r+B->r,A->i-B->i,A->i+B->i,A->r-B->r);

	    V->r=U->r=(g2.r+h2.r-gh.r)/2.0;
		 U->i=(g2.i+h2.i-gh.i)/2.0; V->i=(-U->i);
	    for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=V+Lz1,
					    a=A+1, b=B+Lz1, z=1;
		 z<Lz;
		 ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
							   a++, b--, z++) {
		t=ez->r/ez->i;
		r=Ez[hz].r;
		i=Ez[hz].i; w=r+i*t; c=Product(rh,ih,w*r,w*i);
				     g2=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		r=Ez[gz].r;
		i=Ez[gz].i; w=r+i*t; c=Product(rg,ig,w*r,w*i);
				     h2=Product(c.r,-c.i,u->i+v->i,u->r-v->r);

		gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

		v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		     u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	    }
	    FFT(Lz,Ez,U, 1);
	    FFT(Lz,Ez,V, 1);

	    for (u=U, v=V, z=0; z<Lz; z++) {
		P[z][  x]=(*(u++));
		P[z][Lxx]=(*(v++));
	    }
	}
/*
    y=Ly/2 & x=Lx/2
*/
	for (u=U, a=A, z=0; z<Lz; z++) {
	    *(u++)=P[z][x]; *(a++)=R[z][x];
	}
	FFT(Lz,Ez,U,-1); FFT(Lz,Ez,A,-1);

	U->r=U->r*Nhx2y2*(double)Nhz+Ngx2y2*(double)Ngz*U->i-2.0*A->r*A->i;
	U->i=0.0;
	for (ez=Ez+1, gz=Ngz1, hz=Nhz1, u=U+1, v=U+Lz1,
					a=A+1, b=A+Lz1;
	     u<v;
	     ez++, gz=(gz+Ngz1)%Lz2, hz=(hz+Nhz1)%Lz2, u++, v--,
						       a++, b--) {
	    t=ez->r/ez->i;
	    r=Ez[hz].r;
	    i=Ez[hz].i; w=Nhx2y2*(r+i*t);
			g2=Product(u->r+v->r,u->i-v->i,w*r, w*i);
	    r=Ez[gz].r;
	    i=Ez[gz].i; w=Ngx2y2*(r+i*t);
			h2=Product(w*r,-w*i,u->i+v->i,u->r-v->r);

	    gh=Product(a->r+b->r,a->i-b->i,a->i+b->i,a->r-b->r);

	    v->r=u->r=(g2.r+h2.r-gh.r)/2.0;
		 u->i=(g2.i+h2.i-gh.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*Nhx2y2*Nhz2+Ngx2y2*Ngz2*u->i-2.0*a->r*a->i;
	u->i=0.0;

	FFT(Lz,Ez,U, 1);

	for (u=U, z=0; z<Lz; z++) P[z][x]=(*(u++));

	Replace1(tmp1,Lx,y,Lz,P);
/*
    ----
*/
	Nghx=Ngx-Nhx;
	Nghy=Ngy-Nhy;
	Nghz=Ngz-Nhz;
	L=(double)Lx*(double)Ly*(double)Lz;

	for (z=Dz1; z<=Dz2; z++) {
	    if (z<0) {
		Restore(tmp1,Lx,Ly,Lz+z,P); LNz=L*(double)((z>=Nghz)?Ngz:Nhz+z);
	    }
	    else {
		Restore(tmp1,Lx,Ly,z,P); LNz=L*(double)((z>=Nghz)?Ngz-z:Nhz);
	    }
	    for (x=0; x<Lx; x++) {
		for (u=U, y=0; y<Ly; y++) *(u++)=P[y][x];

		FFT(Ly,Ey,U, 1);

		for (u=U, y=0; y<Ly; y++) P[y][x]=(*(u++));
	    }
	    for (y=Dy1; y<=Dy2; y++) {
		if (y<0) {
		    Px=P[Ly+y]; LNyz=LNz*(double)((y>=Nghy)?Ngy:Nhy+y);
		}
		else {
		    Px=P[y]; LNyz=LNz*(double)((y>=Nghy)?Ngy-y:Nhy);
		}
		FFT(Lx,Ex,Px, 1);

		for (x=Dx1; x<=Dx2; x++) {
		    if (x<0)
			msd=Px[Lx+x].r/(double)((x>=Nghx)?Ngx:Nhx+x);
		    else
			msd=Px[x].r/(double)((x>=Nghx)?Ngx-x:Nhx);

		    (void)printf("%d\t%d\t%d\t%lf\n",
				 x+Dx0,y+Dy0,z+Dz0,
				 sqrt(((msd<0.0)?0.0:msd)/LNyz));
		}
	    }
	}
	return 0;
}
